@@ -1,16 +1,19 @@
-use std::borrow::Cow;
 use std::path::Path;
+use std::sync::Arc;
 
 use anyhow::Context;
 
 use sequoia_openpgp as openpgp;
+use openpgp::cert::raw::RawCertParser;
 use openpgp::Fingerprint;
 use openpgp::KeyHandle;
-use openpgp::Result;
 use openpgp::packet::UserID;
+use openpgp::parse::Parse;
+use openpgp::Result;
 
 use crate::LazyCert;
 use crate::store;
+use store::Certs;
 use store::MergeCerts;
 use store::Store;
 use store::StoreError;
@@ -40,14 +43,17 @@ pub enum AccessMode {
 /// error, the error is ignored, even if it is something other than
 /// [`StoreError::NotFound`].
 ///
-/// Results from the key server are either cached
+/// Results from the key server are either cached when
+/// [`CertStore::flush`] is called (or the [`CertStore`] is dropped)
+/// if there is a writable primary cert-d, or simply dropped
+/// otherwise.
 pub struct CertStore<'a> {
     certd: std::result::Result<store::CertD<'a>, store::Certs<'a>>,
 
     // Read-only backends.
     backends: Vec<(Box<dyn store::Store<'a> + 'a>, AccessMode)>,
 
-    keyserver: Option<Box<dyn store::Store<'a> + 'a>>,
+    keyserver: Option<Box<store::KeyServer<'a>>>,
 }
 
 impl<'a> CertStore<'a> {
@@ -160,18 +166,51 @@ impl<'a> CertStore<'a> {
     pub fn add_keyring<P>(&mut self, path: P) -> Result<&mut Self>
         where P: AsRef<Path>
     {
-        let _path = path.as_ref();
+        self.add_keyrings(std::iter::once(path))?;
         Ok(self)
     }
 
-    /// Adds the specified keybox to the CertStore.
+    /// Adds the specified keyrings to the CertStore.
     ///
-    /// The keybox is added in read-only mode, and its access mode is
-    /// set to `AccessMode::Always`.
-    pub fn add_keybox<P>(&mut self, path: P) -> Result<&mut Self>
-        where P: AsRef<Path>
+    /// The keyrings are added in read-only mode, and their access
+    /// mode is set to `AccessMode::Always`.
+    pub fn add_keyrings<I, P>(&mut self, filenames: I) -> Result<&mut Self>
+    where P: AsRef<Path>,
+          I: IntoIterator<Item=P>,
     {
-        let _path = path.as_ref();
+        let mut keyring = Certs::empty();
+        let mut error = None;
+        for filename in filenames {
+            let filename = filename.as_ref();
+
+            let f = std::fs::File::open(filename)
+                .with_context(|| format!("Open {:?}", filename))?;
+            let parser = RawCertParser::from_reader(f)
+                .with_context(|| format!("Parsing {:?}", filename))?;
+
+            for cert in parser {
+                match cert {
+                    Ok(cert) => {
+                        keyring.update(Arc::new(cert.into()))
+                            .expect("implementation doesn't fail");
+                    }
+                    Err(err) => {
+                        eprint!("Parsing certificate in {:?}: {}",
+                                filename, err);
+                        error = Some(err);
+                    }
+                }
+            }
+        }
+
+        if let Some(err) = error {
+            return Err(err).context("Parsing keyrings");
+        }
+
+        self.add_backend(
+            Box::new(keyring),
+            AccessMode::Always);
+
         Ok(self)
     }
 
@@ -193,10 +232,10 @@ impl<'a> CertStore<'a> {
     /// A key server is treated specially from other backends: any
     /// results that it returns are written to the cert store (if it
     /// is open in read-write mode).
-    pub fn add_keyserver_backend(&mut self, ks: Box<dyn store::Store<'a> + 'a>)
+    pub fn add_keyserver_backend(&mut self, ks: store::KeyServer<'a>)
         -> Result<&mut Self>
     {
-        self.keyserver = Some(ks);
+        self.keyserver = Some(Box::new(ks));
         Ok(self)
     }
 
@@ -271,11 +310,9 @@ macro_rules! forward {
         }
 
         for mode in [AccessMode::Always, AccessMode::OnMiss] {
-            for (backend, m) in $self.backends.iter() {
-                if &mode != m {
-                    continue;
-                }
-
+            for (backend, _) in $self.backends.iter()
+                .filter(|(_, m)| &mode == m)
+            {
                 match backend.$method($($args),*) {
                     Err(err2) => {
                         if let Some(StoreError::NotFound(_))
@@ -300,6 +337,8 @@ macro_rules! forward {
                 }
             }
 
+            // If we found a cert after the AccessMode::Always round,
+            // don't consult the AccessMode::OnMiss backends.
             if mode == AccessMode::Always && ! certs.is_empty() {
                 break;
             }
@@ -344,8 +383,8 @@ macro_rules! forward {
     }}
 }
 
-fn merge<'a, 'b>(mut certs: Vec<Cow<'b, LazyCert<'a>>>)
-    -> Vec<Cow<'b, LazyCert<'a>>>
+fn merge<'a, 'b>(mut certs: Vec<Arc<LazyCert<'a>>>)
+    -> Vec<Arc<LazyCert<'a>>>
 {
     certs.sort_by_key(|cert| cert.fingerprint());
     certs.dedup_by(|a, b| {
@@ -353,13 +392,13 @@ fn merge<'a, 'b>(mut certs: Vec<Cow<'b, LazyCert<'a>>>)
         if a.fingerprint() == b.fingerprint() {
             if let Ok(a2) = a.to_cert() {
                 if let Ok(b2) = b.to_cert() {
-                    *b = Cow::Owned(LazyCert::from(
+                    *b = Arc::new(LazyCert::from(
                         b2.clone()
                             .merge_public(a2.clone())
                             .expect("Same certificate")));
                 } else {
                     // b is invalid, but a is valid.  Just keep a.
-                    *b = Cow::Owned(LazyCert::from(a2.clone()));
+                    *b = Arc::new(LazyCert::from(a2.clone()));
                 }
             } else {
                 // a is invalid.  By returning true, we drop a.
@@ -375,7 +414,7 @@ fn merge<'a, 'b>(mut certs: Vec<Cow<'b, LazyCert<'a>>>)
 
 impl<'a> store::Store<'a> for CertStore<'a> {
     fn lookup_by_cert(&self, kh: &KeyHandle)
-        -> Result<Vec<Cow<LazyCert<'a>>>>
+        -> Result<Vec<Arc<LazyCert<'a>>>>
     {
         let certs = forward!(lookup_by_cert, self, kh, kh)?;
         if certs.is_empty() {
@@ -386,7 +425,7 @@ impl<'a> store::Store<'a> for CertStore<'a> {
     }
 
     fn lookup_by_cert_fpr(&self, fingerprint: &Fingerprint)
-        -> Result<Cow<LazyCert<'a>>>
+        -> Result<Arc<LazyCert<'a>>>
     {
         let certs = forward!(lookup_by_cert_fpr,
                              append:|c| vec![c],
@@ -402,10 +441,10 @@ impl<'a> store::Store<'a> for CertStore<'a> {
         }
     }
 
-    fn lookup_by_key(&self, kh: &KeyHandle)
-        -> Result<Vec<Cow<LazyCert<'a>>>>
+    fn lookup_by_cert_or_subkey(&self, kh: &KeyHandle)
+        -> Result<Vec<Arc<LazyCert<'a>>>>
     {
-        let certs = forward!(lookup_by_key, self, kh, kh)?;
+        let certs = forward!(lookup_by_cert_or_subkey, self, kh, kh)?;
         if certs.is_empty() {
             Err(StoreError::NotFound(kh.clone()).into())
         } else {
@@ -414,7 +453,7 @@ impl<'a> store::Store<'a> for CertStore<'a> {
     }
 
     fn select_userid(&self, query: &UserIDQueryParams, pattern: &str)
-        -> Result<Vec<Cow<LazyCert<'a>>>>
+        -> Result<Vec<Arc<LazyCert<'a>>>>
     {
         let certs = forward!(select_userid, self, pattern, query, pattern)?;
         if certs.is_empty() {
@@ -425,7 +464,7 @@ impl<'a> store::Store<'a> for CertStore<'a> {
     }
 
     fn lookup_by_userid(&self, userid: &UserID)
-        -> Result<Vec<Cow<LazyCert<'a>>>>
+        -> Result<Vec<Arc<LazyCert<'a>>>>
     {
         let certs = forward!(lookup_by_userid, self, userid, userid)?;
         if certs.is_empty() {
@@ -436,7 +475,7 @@ impl<'a> store::Store<'a> for CertStore<'a> {
         }
     }
 
-    fn grep_userid(&self, pattern: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn grep_userid(&self, pattern: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         let certs = forward!(grep_userid, self, pattern, pattern)?;
         if certs.is_empty() {
             Err(StoreError::NoMatches(pattern.to_string()).into())
@@ -445,7 +484,7 @@ impl<'a> store::Store<'a> for CertStore<'a> {
         }
     }
 
-    fn lookup_by_email(&self, email: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_email(&self, email: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         let certs = forward!(lookup_by_email, self, email, email)?;
         if certs.is_empty() {
             Err(StoreError::NoMatches(email.to_string()).into())
@@ -454,7 +493,7 @@ impl<'a> store::Store<'a> for CertStore<'a> {
         }
     }
 
-    fn grep_email(&self, pattern: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn grep_email(&self, pattern: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         let certs = forward!(grep_email, self, pattern, pattern)?;
         if certs.is_empty() {
             Err(StoreError::NoMatches(pattern.to_string()).into())
@@ -464,7 +503,7 @@ impl<'a> store::Store<'a> for CertStore<'a> {
     }
 
     fn lookup_by_email_domain(&self, domain: &str)
-        -> Result<Vec<Cow<LazyCert<'a>>>>
+        -> Result<Vec<Arc<LazyCert<'a>>>>
     {
         let certs = forward!(lookup_by_email_domain, self, domain, domain)?;
         if certs.is_empty() {
@@ -496,7 +535,7 @@ impl<'a> store::Store<'a> for CertStore<'a> {
         Box::new(certs.into_iter())
     }
 
-    fn certs<'b>(&'b self) -> Box<dyn Iterator<Item=Cow<'b, LazyCert<'a>>> + 'b>
+    fn certs<'b>(&'b self) -> Box<dyn Iterator<Item=Arc<LazyCert<'a>>> + 'b>
         where 'a: 'b
     {
         let mut certs = Vec::new();
@@ -530,22 +569,22 @@ impl<'a> store::Store<'a> for CertStore<'a> {
         }
     }
 
-    fn prefetch_some(&mut self, certs: Vec<KeyHandle>) {
+    fn prefetch_some(&mut self, certs: &[KeyHandle]) {
         match self.certd.as_mut() {
-            Ok(certd) => certd.prefetch_some(certs.clone()),
-            Err(in_memory) => in_memory.prefetch_some(certs.clone()),
+            Ok(certd) => certd.prefetch_some(certs),
+            Err(in_memory) => in_memory.prefetch_some(certs),
         };
 
         for (backend, _mode) in self.backends.iter_mut() {
-            backend.prefetch_some(certs.clone());
+            backend.prefetch_some(certs);
         }
     }
 }
 
 impl<'a> store::StoreUpdate<'a> for CertStore<'a> {
-    fn update_by<'ra>(&'ra mut self, cert: Cow<'ra, LazyCert<'a>>,
-                      merge_strategy: &mut dyn MergeCerts<'a, 'ra>)
-        -> Result<Cow<'ra, LazyCert<'a>>>
+    fn update_by(&mut self, cert: Arc<LazyCert<'a>>,
+                 merge_strategy: &mut dyn MergeCerts<'a>)
+        -> Result<Arc<LazyCert<'a>>>
     {
         tracer!(TRACE, "CertStore::update_by");
         match self.certd.as_mut() {
@@ -569,8 +608,9 @@ impl<'a> CertStore<'a> {
     /// work in a write-through manner.
     ///
     /// Note: this is called automatically when the `CertStore` is
-    /// dropped.
-    fn flush(&mut self) -> Result<()> {
+    /// dropped, but calling it explicitly allows for reporting of
+    /// errors.
+    pub fn flush(&mut self) -> Result<()> {
         // Sync the key server's cache to the backing store.
         tracer!(TRACE, "CertStore::flush");
         t!("flushing");
@@ -584,7 +624,7 @@ impl<'a> CertStore<'a> {
             return Ok(());
         };
 
-        let ks = if let Some(ks) = self.keyserver.as_ref() {
+        let ks = if let Some(ks) = self.keyserver.as_mut() {
             ks
         } else {
             // We don't have a key server.  There is clearly nothing
@@ -593,13 +633,11 @@ impl<'a> CertStore<'a> {
             return Ok(());
         };
 
-        let mut count = 0;
+        let mut flushed_certs = Vec::new();
         let mut result = Ok(());
         for c in ks.certs() {
-            count += 1;
-
             let keyid = c.keyid();
-            if let Err(err) = certd.update(c) {
+            if let Err(err) = certd.update(c.clone()) {
                 t!("syncing {} to the cert-d: {}", keyid, err);
                 if result.is_ok() {
                     result = Err(err)
@@ -608,10 +646,17 @@ impl<'a> CertStore<'a> {
                                     keyid)
                         })
                 }
+            } else {
+                flushed_certs.push(c);
             }
         }
 
-        t!("Flushed {} certificates", count);
+        // Now purge the successfully flushed certs from the keyserver
+        // cache, so that we will not try to flush them again next
+        // time we're called.
+        flushed_certs.iter().for_each(|c| ks.delete_from_cache(c));
+
+        t!("Flushed {} certificates", flushed_certs.len());
         result
     }
 }
