@@ -13,20 +13,11 @@
 //! # Examples
 //!
 //! ```rust
-//! use std::borrow::Cow;
+//! use std::sync::Arc;
+//! use sequoia_openpgp::cert::{Cert, CertBuilder};
+//! use sequoia_cert_store::{CertStore, LazyCert, Store, StoreUpdate};
 //!
-//! use sequoia_openpgp as openpgp;
-//! # use openpgp::Result;
-//! use openpgp::cert::Cert;
-//! use openpgp::cert::CertBuilder;
-//!
-//! use sequoia_cert_store as cert_store;
-//! use cert_store::CertStore;
-//! use cert_store::LazyCert;
-//! use cert_store::Store;
-//! use cert_store::StoreUpdate;
-//!
-//! # fn main() -> Result<()> {
+//! # fn main() -> anyhow::Result<()> {
 //! // Create an in-memory certificate store.  To use the default
 //! // on-disk certificate store, use `CertStore::new`.
 //! let mut certs = CertStore::empty();
@@ -39,7 +30,7 @@
 //!
 //! // Insert a certificate.  If using a backing store, it would
 //! // also be written to disk.
-//! certs.update(Cow::Owned(LazyCert::from(cert)))?;
+//! certs.update(Arc::new(LazyCert::from(cert)))?;
 //!
 //! // Make sure it is there.
 //! let cert = certs.lookup_by_cert_fpr(&fpr).expect("present");
@@ -73,21 +64,6 @@ pub use lazy_cert::LazyCert;
 
 const TRACE: bool = false;
 
-/// Prints the error and causes, if any.
-fn print_error_chain(err: &anyhow::Error) {
-    let _ = write_error_chain_into(&mut std::io::stderr(), err);
-}
-
-/// Prints the error and causes, if any.
-fn write_error_chain_into(sink: &mut dyn std::io::Write, err: &anyhow::Error)
-                          -> Result<()> {
-    writeln!(sink, "           {}", err)?;
-    for cause in err.chain().skip(1) {
-        writeln!(sink, "  because: {}", cause)?;
-    }
-    Ok(())
-}
-
 /// Converts an email address to a User ID.
 ///
 /// If the email address is not valid, returns an error.
@@ -106,7 +82,7 @@ fn write_error_chain_into(sink: &mut dyn std::io::Write, err: &anyhow::Error)
 /// [`addr-spec`]: https://www.rfc-editor.org/rfc/rfc2822#section-3.4.1
 pub fn email_to_userid(email: &str) -> Result<UserID> {
     let email_check = UserID::from(format!("<{}>", email));
-    match email_check.email() {
+    match email_check.email2() {
         Ok(Some(email_check)) => {
             if email != email_check {
                 return Err(anyhow::anyhow!(
@@ -145,10 +121,25 @@ pub fn email_to_userid(email: &str) -> Result<UserID> {
 mod tests {
     use super::*;
 
-    use std::borrow::Cow;
+    use std::path::PathBuf;
     use std::str;
+    use std::sync::Arc;
 
     use anyhow::Context;
+    /// Prints the error and causes, if any.
+    pub fn print_error_chain(err: &anyhow::Error) {
+        let _ = write_error_chain_into(&mut std::io::stderr(), err);
+    }
+
+    /// Prints the error and causes, if any.
+    fn write_error_chain_into(sink: &mut dyn std::io::Write, err: &anyhow::Error)
+                              -> Result<()> {
+        writeln!(sink, "           {}", err)?;
+        for cause in err.chain().skip(1) {
+            writeln!(sink, "  because: {}", cause)?;
+        }
+        Ok(())
+    }
 
     use openpgp::Fingerprint;
     use openpgp::KeyHandle;
@@ -161,20 +152,22 @@ mod tests {
     use openpgp_cert_d as cert_d;
 
     use store::Certs;
+    use store::Pep;
     use store::StoreError;
     use store::UserIDQueryParams;
 
-    fn certd_merge(new: cert_d::Data, disk: Option<cert_d::Data>)
-        -> cert_d::Result<cert_d::Data> {
+    fn certd_merge<'a>(new: &'a [u8], disk: Option<&[u8]>)
+        -> cert_d::Result<cert_d::MergeResult<'a>>
+    {
         if let Some(disk) = disk {
             let new = Cert::from_bytes(&new).expect("valid");
             let disk = Cert::from_bytes(&disk).expect("valid");
             let merged = new.merge_public(disk).expect("valid");
             let mut bytes = Vec::new();
             merged.serialize(&mut bytes).expect("valid");
-            Ok(bytes.into_boxed_slice())
+            Ok(bytes.into())
         } else {
-            Ok(new)
+            Ok(cert_d::MergeResult::DataRef(new))
         }
     }
 
@@ -246,7 +239,9 @@ mod tests {
                     Err(err) => {
                         match err.downcast_ref::<StoreError>() {
                             Some(StoreError::NotFound(_)) => (),
-                            _ => panic!("Unexpected failure: {}", err),
+                            _ => panic!("Expected StoreError::NotFound, \
+                                         got: {}",
+                                        err),
                         }
                     },
                 }
@@ -285,36 +280,36 @@ mod tests {
                 }
             }
 
-            // Check lookup_by_key using fingerprints.
-            let got = backend.lookup_by_key(&KeyHandle::from(fpr.clone()))
+            // Check lookup_by_cert_or_subkey using fingerprints.
+            let got = backend.lookup_by_cert_or_subkey(&KeyHandle::from(fpr.clone()))
                 .expect("present");
             assert!(got.into_iter().any(|c| c.fingerprint() == fpr),
-                    "{}, lookup_by_key, with fingerprint, primary",
+                    "{}, lookup_by_cert_or_subkey, with fingerprint, primary",
                     handle.base);
 
             // Look up by subkey and make sure we get cert.
             for sk in cert.keys().subkeys() {
-                let got = backend.lookup_by_key(
+                let got = backend.lookup_by_cert_or_subkey(
                     &KeyHandle::from(sk.fingerprint()))
                     .expect("present");
                 assert!(got.into_iter().any(|c| c.fingerprint() == fpr),
-                        "{}, lookup_by_key, with fingerprint, subkey",
-                        handle.base);
+                        "{}, lookup_by_cert_or_subkey({}), with fingerprint, subkey",
+                        handle.base, sk.fingerprint());
             }
 
 
-            // Check lookup_by_key using keyids.
-            let got = backend.lookup_by_key(&KeyHandle::from(keyid.clone()))
+            // Check lookup_by_cert_or_subkey using keyids.
+            let got = backend.lookup_by_cert_or_subkey(&KeyHandle::from(keyid.clone()))
                 .expect("present");
             assert!(got.into_iter().any(|c| c.fingerprint() == fpr),
-                    "{}, lookup_by_key, with keyid, primary", handle.base);
+                    "{}, lookup_by_cert_or_subkey, with keyid, primary", handle.base);
 
             // Look up by subkey and make sure we get cert.
             for sk in cert.keys().subkeys() {
-                let got = backend.lookup_by_key(&KeyHandle::from(sk.keyid()))
+                let got = backend.lookup_by_cert_or_subkey(&KeyHandle::from(sk.keyid()))
                     .expect("present");
                 assert!(got.into_iter().any(|c| c.fingerprint() == fpr),
-                        "{}, lookup_by_key, with keyid, subkey", handle.base);
+                        "{}, lookup_by_cert_or_subkey, with keyid, subkey", handle.base);
             }
 
 
@@ -391,7 +386,7 @@ mod tests {
                 }
 
                 // Search by exact email.
-                let email = if let Ok(Some(email)) = userid.email() {
+                let email = if let Ok(Some(email)) = userid.email2() {
                     email
                 } else {
                     // No email address.
@@ -546,7 +541,7 @@ mod tests {
 
         // alice and alice2 share a subkey.
         assert_eq!(
-            sort_vec(backend.lookup_by_key(
+            sort_vec(backend.lookup_by_cert_or_subkey(
                 &"5989D7BE9908AE24799DF6CFBE678043781349F1"
                     .parse::<KeyHandle>().expect("valid"))
                 .expect("present")
@@ -564,7 +559,7 @@ mod tests {
 
         // ed's primary is also a subkey on the same certificate.
         assert_eq!(
-            sort_vec(backend.lookup_by_key(
+            sort_vec(backend.lookup_by_cert_or_subkey(
                 &"0C346B2B6241263F64E9C7CF1EA300797258A74E"
                     .parse::<KeyHandle>().expect("valid"))
                 .expect("present")
@@ -583,7 +578,7 @@ mod tests {
         // but the backend is not supposed to check that.  (That
         // subkey is bound to carol.)
         assert_eq!(
-            sort_vec(backend.lookup_by_key(
+            sort_vec(backend.lookup_by_cert_or_subkey(
                 &"CD22D4BD99FF10FDA11A83D4213DCB92C95346CE"
                     .parse::<KeyHandle>().expect("valid"))
                 .expect("present")
@@ -614,11 +609,11 @@ mod tests {
             }
         }
 
-        match backend.lookup_by_key(
+        match backend.lookup_by_cert_or_subkey(
             &"0123 4567 89AB CDEF 0123 4567 89AB CDEF"
                 .parse::<KeyHandle>().expect("valid"))
         {
-            Ok(certs) => panic!("lookup_by_key(not present) -> {}",
+            Ok(certs) => panic!("lookup_by_cert_or_subkey(not present) -> {}",
                                 certs
                                     .into_iter()
                                     .map(|c| c.fingerprint().to_string())
@@ -633,7 +628,7 @@ mod tests {
         }
 
         assert!(
-            backend.lookup_by_key(
+            backend.lookup_by_cert_or_subkey(
                 &"0123 4567 89AB CDEF 0123 4567 89AB CDEF"
                     .parse::<KeyHandle>().expect("valid"))
                 .is_err());
@@ -759,7 +754,7 @@ mod tests {
                 .expect(&format!("{}", cert.base));
 
             certd
-                .insert(bytes.into_boxed_slice(), certd_merge)
+                .insert_data(&bytes, false, certd_merge)
                 .with_context(|| {
                     format!("{} ({})", cert.base, cert.fingerprint)
                 })
@@ -800,7 +795,7 @@ mod tests {
                 .expect(&format!("{}", cert.base));
 
             certd
-                .insert(bytes.into_boxed_slice(), certd_merge)
+                .insert_data(&bytes, false, certd_merge)
                 .with_context(|| {
                     format!("{} ({})", cert.base, cert.fingerprint)
                 })
@@ -844,7 +839,7 @@ mod tests {
                 .expect(&format!("{}", cert.base));
 
             certd
-                .insert(bytes.into_boxed_slice(), certd_merge)
+                .insert_data(&bytes, false, certd_merge)
                 .with_context(|| {
                     format!("{} ({})", cert.base, cert.fingerprint)
                 })
@@ -910,7 +905,7 @@ mod tests {
                 .expect(&format!("{}", cert.base));
 
             certd
-                .insert(bytes.into_boxed_slice(), certd_merge)
+                .insert_data(&bytes, false, certd_merge)
                 .with_context(|| {
                     format!("{} ({})", cert.base, cert.fingerprint)
                 })
@@ -949,6 +944,46 @@ mod tests {
         Ok(())
     }
 
+    #[test]
+    fn keyrings() -> Result<()> {
+        let mut cert_store = CertStore::empty();
+
+        let mut base = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
+        base.push("tests");
+
+        cert_store.add_keyrings(
+            keyring::certs.iter().map(|c| {
+                PathBuf::from(&base).join(c.filename)
+            }))?;
+
+        test_backend(cert_store);
+
+        Ok(())
+    }
+
+    #[test]
+    fn pep() -> Result<()> {
+        use std::io::Read;
+
+        assert_eq!(keyring::certs.len(), 12);
+
+        let mut bytes = Vec::new();
+        for cert in keyring::certs.iter() {
+            let binary = cert.bytes();
+            let mut reader = openpgp::armor::Reader::from_bytes(
+                &binary,
+                openpgp::armor::ReaderMode::VeryTolerant);
+            reader.read_to_end(&mut bytes)
+                .expect(&format!("{}", cert.base));
+        }
+
+        let mut backend = Pep::from_bytes(&bytes).expect("valid");
+        backend.prefetch_all();
+        test_backend(backend);
+
+        Ok(())
+    }
+
     // Make sure that when we update a certificate, we are able to
     // find any new components and we are still able to find the old
     // components.
@@ -1017,20 +1052,20 @@ mod tests {
             assert_eq!(r.len(), 1);
             assert_eq!(r[0].to_cert().expect("ok"), cert);
 
-            let r = backend.lookup_by_key(&signing_fpr).unwrap();
+            let r = backend.lookup_by_cert_or_subkey(&signing_fpr).unwrap();
             assert_eq!(r.len(), 1);
             assert_eq!(r[0].to_cert().expect("ok"), cert);
 
-            let r = backend.lookup_by_key(&auth_fpr).unwrap();
+            let r = backend.lookup_by_cert_or_subkey(&auth_fpr).unwrap();
             assert_eq!(r.len(), 1);
             assert_eq!(r[0].to_cert().expect("ok"), cert);
 
             if have_enc {
-                let r = backend.lookup_by_key(&encryption_fpr).unwrap();
+                let r = backend.lookup_by_cert_or_subkey(&encryption_fpr).unwrap();
                 assert_eq!(r.len(), 1);
                 assert_eq!(r[0].to_cert().expect("ok"), cert);
             } else {
-                assert!(backend.lookup_by_key(&encryption_fpr).is_err());
+                assert!(backend.lookup_by_cert_or_subkey(&encryption_fpr).is_err());
             }
 
             let r = backend.lookup_by_userid(
@@ -1050,20 +1085,20 @@ mod tests {
                 assert_eq!(r.len(), 1);
                 assert_eq!(r[0].to_cert().expect("ok"), cert);
             } else {
-                assert!(backend.lookup_by_key(&encryption_fpr).is_err());
+                assert!(backend.lookup_by_cert_or_subkey(&encryption_fpr).is_err());
             }
         };
 
         // Insert the signing certificate.
-        backend.update(Cow::Owned(LazyCert::from(signing_cert.clone())))
+        backend.update(Arc::new(LazyCert::from(signing_cert.clone())))
             .expect("ok");
         check(&backend, false, &signing_cert);
 
-        backend.update(Cow::Owned(LazyCert::from(encryption_cert.clone())))
+        backend.update(Arc::new(LazyCert::from(encryption_cert.clone())))
             .expect("ok");
         check(&backend, true, &merged_cert);
 
-        backend.update(Cow::Owned(LazyCert::from(signing_cert.clone())))
+        backend.update(Arc::new(LazyCert::from(signing_cert.clone())))
             .expect("ok");
         check(&backend, true, &merged_cert);
 
@@ -1084,5 +1119,12 @@ mod tests {
         let certs = Certs::empty();
         test_store_update(certs)
     }
+
+    // Test StoreUpdate::update for Pep.
+    #[test]
+    fn test_store_update_pep() -> Result<()> {
+        let certs = Pep::empty()?;
+        test_store_update(certs)
+    }
 }
 
