@@ -1,5 +1,5 @@
-use std::borrow::Cow;
 use std::str;
+use std::sync::Arc;
 
 use anyhow::Context;
 
@@ -11,6 +11,8 @@ use openpgp::cert::Cert;
 use openpgp::cert::ValidCert;
 use openpgp::packet::UserID;
 
+pub use openpgp_cert_d;
+
 mod userid_index;
 pub use userid_index::UserIDIndex;
 
@@ -23,6 +25,9 @@ pub use certs::Certs;
 pub mod keyserver;
 pub use keyserver::KeyServer;
 
+pub mod pep;
+pub use pep::Pep;
+
 use super::TRACE;
 
 use crate::LazyCert;
@@ -238,7 +243,7 @@ impl UserIDQueryParams {
     /// If it is valid, returns the normalized email address.
     pub fn is_email(email: &str) -> Result<String> {
         let email_check = UserID::from(format!("<{}>", email));
-        match email_check.email() {
+        match email_check.email2() {
             Ok(Some(email_check)) => {
                 if email != email_check {
                     return Err(StoreError::InvalidEmail(
@@ -313,7 +318,7 @@ pub enum StoreError {
 /// and `LookupFailed` will use the same transaction id.
 #[non_exhaustive]
 #[derive(Debug)]
-pub enum StatusUpdate<'a, 'c: 'rc, 'rc> {
+pub enum StatusUpdate<'a, 'c> {
     /// Sent when a lookup is starting.
     ///
     /// usize is the transaction id.
@@ -356,7 +361,7 @@ pub enum StatusUpdate<'a, 'c: 'rc, 'rc> {
     /// The last `&str` is a short human-readable message describing
     /// what happened, e.g., "found in cache", etc.
     LookupFinished(usize, &'a str, &'a KeyHandle,
-                   &'a [Cow<'rc, LazyCert<'c>>], Option<&'a str>),
+                   &'a [Arc<LazyCert<'c>>], Option<&'a str>),
 
     /// Sent when a lookup has failed.
     ///
@@ -414,7 +419,7 @@ pub enum StatusUpdate<'a, 'c: 'rc, 'rc> {
     /// The last `&str` is a short human-readable message describing
     /// what happened, e.g., "found in cache", "found 5 matching
     /// certificates", etc.
-    SearchFinished(usize, &'a str, &'a str, &'a [Cow<'rc, LazyCert<'c>>],
+    SearchFinished(usize, &'a str, &'a str, &'a [Arc<LazyCert<'c>>],
                    Option<&'a str>),
 
     /// Sent whenever something has been lookup successfully.
@@ -433,7 +438,7 @@ pub enum StatusUpdate<'a, 'c: 'rc, 'rc> {
     SearchFailed(usize, &'a str, &'a str, Option<&'a anyhow::Error>),
 }
 
-impl<'a, 'c: 'rc, 'rc> std::fmt::Display for StatusUpdate<'a, 'c, 'rc> {
+impl<'a, 'c> std::fmt::Display for StatusUpdate<'a, 'c> {
     fn fmt(&self, fmt: &mut std::fmt::Formatter<'_>)
         -> std::result::Result<(), std::fmt::Error>
     {
@@ -532,7 +537,7 @@ pub trait Store<'a> {
     ///
     /// The caller may assume that looking up a fingerprint returns at
     /// most one certificate.
-    fn lookup_by_cert(&self, kh: &KeyHandle) -> Result<Vec<Cow<LazyCert<'a>>>>;
+    fn lookup_by_cert(&self, kh: &KeyHandle) -> Result<Vec<Arc<LazyCert<'a>>>>;
 
     /// Returns the certificate with the specified fingerprint, if any.
     ///
@@ -541,7 +546,7 @@ pub trait Store<'a> {
     /// The default implementation is implemented in terms of
     /// [`Store::lookup_by_cert`].
     fn lookup_by_cert_fpr(&self, fingerprint: &Fingerprint)
-        -> Result<Cow<LazyCert<'a>>>
+        -> Result<Arc<LazyCert<'a>>>
     {
         let kh = KeyHandle::from(fingerprint.clone());
 
@@ -561,17 +566,17 @@ pub trait Store<'a> {
     /// Note: even if you pass a fingerprint, this may return multiple
     /// certificates as the same subkey may be attached to multiple
     /// certificates.
-    fn lookup_by_key(&self, kh: &KeyHandle) -> Result<Vec<Cow<LazyCert<'a>>>>;
+    fn lookup_by_cert_or_subkey(&self, kh: &KeyHandle) -> Result<Vec<Arc<LazyCert<'a>>>>;
 
     /// Returns certificates that have a User ID matching the
     /// specified pattern according to the query parameters.
     fn select_userid(&self, query: &UserIDQueryParams, pattern: &str)
-        -> Result<Vec<Cow<LazyCert<'a>>>>;
+        -> Result<Vec<Arc<LazyCert<'a>>>>;
 
     /// Performs an exact match on the User ID.
     ///
     /// The pattern is anchored, and the match is case sensitive.
-    fn lookup_by_userid(&self, userid: &UserID) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_userid(&self, userid: &UserID) -> Result<Vec<Arc<LazyCert<'a>>>> {
         self.select_userid(
             &UserIDQueryParams::new()
                 .set_email(false)
@@ -585,7 +590,7 @@ pub trait Store<'a> {
     ///
     /// The pattern is not anchored, and it is matched case
     /// insensitively.
-    fn grep_userid(&self, pattern: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn grep_userid(&self, pattern: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         self.select_userid(
             &UserIDQueryParams::new()
                 .set_email(false)
@@ -595,13 +600,13 @@ pub trait Store<'a> {
             pattern)
     }
 
-    /// Returns certificates that have User ID with the specified
+    /// Returns certificates that have a User ID with the specified
     /// email address.
     ///
     /// The pattern is interpreted as an email address.  It is first
     /// normalized, and then matched against the normalized email
     /// address, it is anchored, and the match is case sensitive.
-    fn lookup_by_email(&self, email: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_email(&self, email: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         let userid = crate::email_to_userid(&email)?;
         let email = userid.email_normalized()?.expect("have one");
 
@@ -620,7 +625,7 @@ pub trait Store<'a> {
     /// The pattern is matched against the normalized email address,
     /// it is not anchored, and it is matched case insensitively.  The
     /// pattern itself is *not* normalized.
-    fn grep_email(&self, pattern: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn grep_email(&self, pattern: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         self.select_userid(
             &UserIDQueryParams::new()
                 .set_email(true)
@@ -641,7 +646,7 @@ pub trait Store<'a> {
     /// not start with an `@`.  This does not match subdomains.  That
     /// is, it will match `alice@foo.bar.com` when searching for
     /// `bar.com`.
-    fn lookup_by_email_domain(&self, domain: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_email_domain(&self, domain: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         let localpart = "localpart";
         let email = format!("{}@{}", localpart, domain);
         let userid = crate::email_to_userid(&email)?;
@@ -673,7 +678,7 @@ pub trait Store<'a> {
     /// [`Store::fingerprints`] and [`Store::lookup_by_cert_fpr`].  Many backends
     /// will be able to do this more efficiently.
     fn certs<'b>(&'b self)
-        -> Box<dyn Iterator<Item=Cow<'b, LazyCert<'a>>> + 'b>
+        -> Box<dyn Iterator<Item=Arc<LazyCert<'a>>> + 'b>
         where 'a: 'b
     {
         Box::new(self.fingerprints()
@@ -706,7 +711,7 @@ pub trait Store<'a> {
     ///
     /// Errors should be silently ignored and propagated when the
     /// operation in question is executed directly.
-    fn prefetch_some(&mut self, certs: Vec<KeyHandle>) {
+    fn prefetch_some(&mut self, certs: &[KeyHandle]) {
         let _ = certs;
     }
 }
@@ -717,43 +722,43 @@ pub trait Store<'a> {
 impl<'a: 't, 't, T> Store<'a> for Box<T>
 where T: Store<'a> + ?Sized + 't
 {
-    fn lookup_by_cert(&self, kh: &KeyHandle) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_cert(&self, kh: &KeyHandle) -> Result<Vec<Arc<LazyCert<'a>>>> {
         self.as_ref().lookup_by_cert(kh)
     }
 
     fn lookup_by_cert_fpr(&self, fingerprint: &Fingerprint)
-        -> Result<Cow<LazyCert<'a>>>
+        -> Result<Arc<LazyCert<'a>>>
     {
         self.as_ref().lookup_by_cert_fpr(fingerprint)
     }
 
-    fn lookup_by_key(&self, kh: &KeyHandle) -> Result<Vec<Cow<LazyCert<'a>>>> {
-        self.as_ref().lookup_by_key(kh)
+    fn lookup_by_cert_or_subkey(&self, kh: &KeyHandle) -> Result<Vec<Arc<LazyCert<'a>>>> {
+        self.as_ref().lookup_by_cert_or_subkey(kh)
     }
 
     fn select_userid(&self, query: &UserIDQueryParams, pattern: &str)
-        -> Result<Vec<Cow<LazyCert<'a>>>>
+        -> Result<Vec<Arc<LazyCert<'a>>>>
     {
         self.as_ref().select_userid(query, pattern)
     }
 
-    fn lookup_by_userid(&self, userid: &UserID) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_userid(&self, userid: &UserID) -> Result<Vec<Arc<LazyCert<'a>>>> {
         self.as_ref().lookup_by_userid(userid)
     }
 
-    fn grep_userid(&self, pattern: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn grep_userid(&self, pattern: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         self.as_ref().grep_userid(pattern)
     }
 
-    fn lookup_by_email(&self, email: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_email(&self, email: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         self.as_ref().lookup_by_email(email)
     }
 
-    fn grep_email(&self, pattern: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn grep_email(&self, pattern: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         self.as_ref().grep_email(pattern)
     }
 
-    fn lookup_by_email_domain(&self, domain: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_email_domain(&self, domain: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         self.as_ref().lookup_by_email_domain(domain)
     }
 
@@ -762,7 +767,7 @@ where T: Store<'a> + ?Sized + 't
     }
 
     fn certs<'b>(&'b self)
-        -> Box<dyn Iterator<Item=Cow<'b, LazyCert<'a>>> + 'b>
+        -> Box<dyn Iterator<Item=Arc<LazyCert<'a>>> + 'b>
         where 'a: 'b
     {
         self.as_ref().certs()
@@ -772,7 +777,7 @@ where T: Store<'a> + ?Sized + 't
         self.as_ref().prefetch_all()
     }
 
-    fn prefetch_some(&mut self, certs: Vec<KeyHandle>) {
+    fn prefetch_some(&mut self, certs: &[KeyHandle]) {
         self.as_ref().prefetch_some(certs)
     }
 }
@@ -780,43 +785,43 @@ where T: Store<'a> + ?Sized + 't
 impl<'a: 't, 't, T> Store<'a> for &'t T
 where T: Store<'a> + ?Sized
 {
-    fn lookup_by_cert(&self, kh: &KeyHandle) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_cert(&self, kh: &KeyHandle) -> Result<Vec<Arc<LazyCert<'a>>>> {
         (*self).lookup_by_cert(kh)
     }
 
     fn lookup_by_cert_fpr(&self, fingerprint: &Fingerprint)
-        -> Result<Cow<LazyCert<'a>>>
+        -> Result<Arc<LazyCert<'a>>>
     {
         (*self).lookup_by_cert_fpr(fingerprint)
     }
 
-    fn lookup_by_key(&self, kh: &KeyHandle) -> Result<Vec<Cow<LazyCert<'a>>>> {
-        (*self).lookup_by_key(kh)
+    fn lookup_by_cert_or_subkey(&self, kh: &KeyHandle) -> Result<Vec<Arc<LazyCert<'a>>>> {
+        (*self).lookup_by_cert_or_subkey(kh)
     }
 
     fn select_userid(&self, query: &UserIDQueryParams, pattern: &str)
-        -> Result<Vec<Cow<LazyCert<'a>>>>
+        -> Result<Vec<Arc<LazyCert<'a>>>>
     {
         (*self).select_userid(query, pattern)
     }
 
-    fn lookup_by_userid(&self, userid: &UserID) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_userid(&self, userid: &UserID) -> Result<Vec<Arc<LazyCert<'a>>>> {
         (*self).lookup_by_userid(userid)
     }
 
-    fn grep_userid(&self, pattern: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn grep_userid(&self, pattern: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         (*self).grep_userid(pattern)
     }
 
-    fn lookup_by_email(&self, email: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_email(&self, email: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         (*self).lookup_by_email(email)
     }
 
-    fn grep_email(&self, pattern: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn grep_email(&self, pattern: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         (*self).grep_email(pattern)
     }
 
-    fn lookup_by_email_domain(&self, domain: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_email_domain(&self, domain: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         (*self).lookup_by_email_domain(domain)
     }
 
@@ -825,7 +830,7 @@ where T: Store<'a> + ?Sized
     }
 
     fn certs<'b>(&'b self)
-        -> Box<dyn Iterator<Item=Cow<'b, LazyCert<'a>>> + 'b>
+        -> Box<dyn Iterator<Item=Arc<LazyCert<'a>>> + 'b>
         where 'a: 'b
     {
         (*self).certs()
@@ -835,43 +840,43 @@ where T: Store<'a> + ?Sized
 impl<'a: 't, 't, T> Store<'a> for &'t mut T
 where T: Store<'a> + ?Sized
 {
-    fn lookup_by_cert(&self, kh: &KeyHandle) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_cert(&self, kh: &KeyHandle) -> Result<Vec<Arc<LazyCert<'a>>>> {
         (**self).lookup_by_cert(kh)
     }
 
     fn lookup_by_cert_fpr(&self, fingerprint: &Fingerprint)
-        -> Result<Cow<LazyCert<'a>>>
+        -> Result<Arc<LazyCert<'a>>>
     {
         (**self).lookup_by_cert_fpr(fingerprint)
     }
 
-    fn lookup_by_key(&self, kh: &KeyHandle) -> Result<Vec<Cow<LazyCert<'a>>>> {
-        (**self).lookup_by_key(kh)
+    fn lookup_by_cert_or_subkey(&self, kh: &KeyHandle) -> Result<Vec<Arc<LazyCert<'a>>>> {
+        (**self).lookup_by_cert_or_subkey(kh)
     }
 
     fn select_userid(&self, query: &UserIDQueryParams, pattern: &str)
-        -> Result<Vec<Cow<LazyCert<'a>>>>
+        -> Result<Vec<Arc<LazyCert<'a>>>>
     {
         (**self).select_userid(query, pattern)
     }
 
-    fn lookup_by_userid(&self, userid: &UserID) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_userid(&self, userid: &UserID) -> Result<Vec<Arc<LazyCert<'a>>>> {
         (**self).lookup_by_userid(userid)
     }
 
-    fn grep_userid(&self, pattern: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn grep_userid(&self, pattern: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         (**self).grep_userid(pattern)
     }
 
-    fn lookup_by_email(&self, email: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_email(&self, email: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         (**self).lookup_by_email(email)
     }
 
-    fn grep_email(&self, pattern: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn grep_email(&self, pattern: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         (**self).grep_email(pattern)
     }
 
-    fn lookup_by_email_domain(&self, domain: &str) -> Result<Vec<Cow<LazyCert<'a>>>> {
+    fn lookup_by_email_domain(&self, domain: &str) -> Result<Vec<Arc<LazyCert<'a>>>> {
         (**self).lookup_by_email_domain(domain)
     }
 
@@ -880,7 +885,7 @@ where T: Store<'a> + ?Sized
     }
 
     fn certs<'b>(&'b self)
-        -> Box<dyn Iterator<Item=Cow<'b, LazyCert<'a>>> + 'b>
+        -> Box<dyn Iterator<Item=Arc<LazyCert<'a>>> + 'b>
         where 'a: 'b
     {
         (**self).certs()
@@ -890,7 +895,7 @@ where T: Store<'a> + ?Sized
         (**self).prefetch_all()
     }
 
-    fn prefetch_some(&mut self, certs: Vec<KeyHandle>) {
+    fn prefetch_some(&mut self, certs: &[KeyHandle]) {
         (**self).prefetch_some(certs)
     }
 }
@@ -899,34 +904,37 @@ where T: Store<'a> + ?Sized
 ///
 /// This is primarily useful as the `merge_strategy` callback to
 /// [`StoreUpdate::update_by`].
-pub trait MergeCerts<'a: 'ra, 'ra> {
+pub trait MergeCerts<'a> {
     /// Merges two certificates.
     ///
     /// This is primarily useful as the `merge_strategy` callback to
     /// [`StoreUpdate::update_by`].
     ///
     /// The default implementation merges the two certificates using
-    /// [`Cert::merge_public`].  When a variant of a packet is present
-    /// in the on-disk version and the new version, the variant in the
-    /// new version is preferred.  This can be the case with signature
-    /// packets, for instance, when the unhashed subpacket areas
-    /// differ, but the signatures are otherwise the same.
-    fn merge<'b, 'rb>(&mut self,
-                      new: Cow<'ra, LazyCert<'a>>,
-                      disk: Option<Cow<'rb, LazyCert<'b>>>)
-                      -> Result<Cow<'ra, LazyCert<'a>>>
+    /// [`Cert::merge_public`].  This means that any secret key
+    /// material in `disk` is preserved, any secret key material in
+    /// `new` is ignored, and unhashed subpacket areas are merged.
+    fn merge_public<'b>(&mut self,
+                        new: Arc<LazyCert<'a>>,
+                        disk: Option<Arc<LazyCert<'b>>>)
+        -> Result<Arc<LazyCert<'a>>>
     {
         if let Some(disk) = disk {
-            let merged = new.into_owned().into_cert()?
-                .merge_public(disk.as_cert()?)?;
-            Ok(Cow::Owned(LazyCert::from(merged)))
+            let merged = disk.to_cert()?.clone()
+                .merge_public(new.to_cert()?.clone())?;
+            Ok(Arc::new(LazyCert::from(merged)))
         } else {
-            Ok(new)
+            if new.is_tsk() {
+                Ok(Arc::new(LazyCert::from(new.to_cert()?.clone()
+                                           .strip_secret_key_material())))
+            } else {
+                Ok(new)
+            }
         }
     }
 }
 
-impl<'a: 'ra, 'ra> MergeCerts<'a, 'ra> for () {
+impl<'a> MergeCerts<'a> for () {
 }
 
 /// Provides an interface to update a backing store.
@@ -935,7 +943,7 @@ pub trait StoreUpdate<'a>: Store<'a> {
     ///
     /// This uses the default implementation of [`MergeCerts`] to
     /// merge the certificate with any existing certificate.
-    fn update(&mut self, cert: Cow<LazyCert<'a>>) -> Result<()> {
+    fn update(&mut self, cert: Arc<LazyCert<'a>>) -> Result<()> {
         self.update_by(cert, &mut ())?;
 
         Ok(())
@@ -955,21 +963,21 @@ pub trait StoreUpdate<'a>: Store<'a> {
     ///
     /// To use the default merge strategy, either call
     /// [`StoreUpdate::update`] directly, or pass `&mut ()`.
-    fn update_by<'ra>(&'ra mut self, cert: Cow<'ra, LazyCert<'a>>,
-                      merge_strategy: &mut dyn MergeCerts<'a, 'ra>)
-        -> Result<Cow<'ra, LazyCert<'a>>>;
+    fn update_by(&mut self, cert: Arc<LazyCert<'a>>,
+                 merge_strategy: &mut dyn MergeCerts<'a>)
+        -> Result<Arc<LazyCert<'a>>>;
 }
 
 impl<'a: 't, 't, T> StoreUpdate<'a> for Box<T>
 where T: StoreUpdate<'a> + ?Sized + 't
 {
-    fn update(&mut self, cert: Cow<LazyCert<'a>>) -> Result<()> {
+    fn update(&mut self, cert: Arc<LazyCert<'a>>) -> Result<()> {
         self.as_mut().update(cert)
     }
 
-    fn update_by<'ra>(&'ra mut self, cert: Cow<'ra, LazyCert<'a>>,
-                      merge_strategy: &mut dyn MergeCerts<'a, 'ra>)
-        -> Result<Cow<'ra, LazyCert<'a>>>
+    fn update_by(&mut self, cert: Arc<LazyCert<'a>>,
+                      merge_strategy: &mut dyn MergeCerts<'a>)
+        -> Result<Arc<LazyCert<'a>>>
     {
         self.as_mut().update_by(cert, merge_strategy)
     }
@@ -978,13 +986,13 @@ where T: StoreUpdate<'a> + ?Sized + 't
 impl<'a: 't, 't, T> StoreUpdate<'a> for &'t mut T
 where T: StoreUpdate<'a> + ?Sized
 {
-    fn update(&mut self, cert: Cow<LazyCert<'a>>) -> Result<()> {
+    fn update(&mut self, cert: Arc<LazyCert<'a>>) -> Result<()> {
         (*self).update(cert)
     }
 
-    fn update_by<'ra>(&'ra mut self, cert: Cow<'ra, LazyCert<'a>>,
-                      merge_strategy: &mut dyn MergeCerts<'a, 'ra>)
-        -> Result<Cow<'ra, LazyCert<'a>>>
+    fn update_by(&mut self, cert: Arc<LazyCert<'a>>,
+                      merge_strategy: &mut dyn MergeCerts<'a>)
+        -> Result<Arc<LazyCert<'a>>>
     {
         (*self).update_by(cert, merge_strategy)
     }
@@ -1026,7 +1034,7 @@ impl MergePublicCollectStats {
     }
 }
 
-impl<'a: 'ra, 'ra> MergeCerts<'a, 'ra> for MergePublicCollectStats {
+impl<'a> MergeCerts<'a> for MergePublicCollectStats {
     /// Merges two certificates.
     ///
     /// This is primarily useful as the `merge_strategy` callback to
@@ -1040,7 +1048,7 @@ impl<'a: 'ra, 'ra> MergeCerts<'a, 'ra> for MergePublicCollectStats {
     ///
     /// ```rust
     /// use std::any::Any;
-    /// use std::borrow::Cow;
+    /// use std::sync::Arc;
     ///
     /// use sequoia_openpgp as openpgp;
     /// # use openpgp::Result;
@@ -1060,55 +1068,61 @@ impl<'a: 'ra, 'ra> MergeCerts<'a, 'ra> for MergePublicCollectStats {
     ///
     /// let mut stats = MergePublicCollectStats::new();
     ///
-    /// certs.update_by(Cow::Owned(LazyCert::from(cert)), &mut stats)
+    /// certs.update_by(Arc::new(LazyCert::from(cert)), &mut stats)
     ///         .expect("valid");
     ///
     /// assert_eq!(stats.new, 1);
     /// # Ok(()) }
     /// ```
-    fn merge<'b, 'rb>(&mut self,
-                      new: Cow<'ra, LazyCert<'a>>,
-                      disk: Option<Cow<'rb, LazyCert<'b>>>)
-        -> Result<Cow<'ra, LazyCert<'a>>>
+    fn merge_public<'b, 'rb>(&mut self,
+                             new: Arc<LazyCert<'a>>,
+                             disk: Option<Arc<LazyCert<'b>>>)
+        -> Result<Arc<LazyCert<'a>>>
     {
         let disk = if let Some(disk) = disk {
             disk
         } else {
             self.new += 1;
-            return Ok(new);
+
+            if new.is_tsk() {
+                return Ok(Arc::new(LazyCert::from(
+                    new.to_cert()?.clone().strip_secret_key_material())))
+            } else {
+                return Ok(new);
+            }
         };
 
         let fpr = new.fingerprint();
 
-        let disk = disk.into_owned().as_cert()
+        let disk = disk.to_cert()
             .with_context(|| {
                 format!("Parsing {} as returned from the cert directory", fpr)
-            })?;
+            })?
+            .clone();
 
-        let new = new.into_owned().as_cert()
+        let new = new.to_cert()
             .with_context(|| {
                 format!("Parsing {} as being inserted into \
                          the cert directory",
                         fpr)
-            })?;
+            })?
+            .clone();
 
-        // If the on-disk version has secrets, we
-        // preserve them.
-        let disk_packets = disk.into_packets();
-
-        match new.insert_packets2(disk_packets) {
-            Ok((merged, changed)) => {
-                if changed {
+        if disk == new {
+            self.unchanged += 1;
+            Ok(Arc::new(LazyCert::from(new)))
+        } else {
+            // If the on-disk version has secrets, we preserve them.
+            // If new has secrets, we ignore them.
+            match disk.merge_public(new) {
+                Ok(merged) => {
                     self.updated += 1;
-                } else {
-                    self.unchanged += 1;
+                    Ok(Arc::new(LazyCert::from(merged)))
+                }
+                Err(err) => {
+                    self.errors += 1;
+                    Err(err.into())
                 }
-
-                Ok(Cow::Owned(LazyCert::from(merged)))
-            }
-            Err(err) => {
-                self.errors += 1;
-                Err(err.into())
             }
         }
     }
@@ -1236,7 +1250,6 @@ mod tests {
     // Check that MergePublicCollectStats works as advertised.
     #[test]
     fn store_update_merge_public_collect_stats() {
-        use std::borrow::Cow;
         use std::collections::HashSet;
 
         use openpgp::Cert;
@@ -1258,7 +1271,7 @@ mod tests {
             let fpr = cert.fingerprint();
             seen.insert(fpr.clone());
 
-            certs.update_by(Cow::Owned(LazyCert::from(cert)), &mut stats)
+            certs.update_by(Arc::new(LazyCert::from(cert)), &mut stats)
                 .expect("valid");
 
             eprintln!("After inserting {} ({}), stats: {:?}",
@@ -1277,7 +1290,7 @@ mod tests {
             let cert = Cert::from_bytes(&cert.bytes()).expect("valid");
             let fpr = cert.fingerprint();
 
-            certs.update_by(Cow::Owned(LazyCert::from(cert)), &mut stats)
+            certs.update_by(Arc::new(LazyCert::from(cert)), &mut stats)
                 .expect("valid");
 
             eprintln!("After reinserting {} ({}), stats: {:?}",
